"""The core of the mlpype library: run a standardised ML experiment with the given parameters."""
from argparse import ArgumentParser
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

from mlpype.base.data import DataCatalog, DataSet
from mlpype.base.evaluate import BaseEvaluator
from mlpype.base.evaluate.plot import BasePlotter
from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_pipeline
from mlpype.base.logger import ExperimentLogger
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline
from mlpype.base.pipeline.type_checker import TypeChecker, TypeCheckerPipe
from mlpype.base.serialiser import JoblibSerialiser, Serialiser
from mlpype.base.utils.parsing import get_args_for_prefix


class Experiment:
    """The core of the mlpype library: run a standardised ML experiment with the given parameters.

    This combines all the components of the mlpype library to run a standardised ML experiment.
    DataCatalogs will be initialised (`read`) in the beginning of the run. After this,
    the Pipeline will be fitted on the train data, and the Model will be trained on the
    transformed train data. After this, the Model will be evaluated on all data.

    All metrics, parameters, model and pipeline artifacts, and plots will be logged. Finally,
    the performance metrics will be returned.

    A trained/logged Experiment can be reloaded using the Inferencer class and its `from_folder` method.

    It is highly recommended to use one of the class methods to initialise this object:
            - from_dictionary: If you've already got your parameters in the right form:
                - `model__<arg>` for model parameters
                - `pipeline__<pipe_name>__<arg> for pipeline parameters.
    """

    def __init__(
        self,
        data_sources: Dict[str, Union[DataCatalog, DataSet]],
        model: Model,
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        type_checker_classes: Optional[List[Type[TypeChecker]]] = None,
        input_type_checker: Optional[TypeCheckerPipe] = None,
        output_type_checker: Optional[TypeCheckerPipe] = None,
        serialiser: Optional[Serialiser] = None,
        output_folder: Union[Path, str] = "outputs",
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        plots: Optional[List[BasePlotter]] = None,
    ):
        """The core of the mlpype library: run a standardised ML experiment with the given parameters.

        It is highly recommended to use one of the class methods to initialise this object:
            - from_dictionary: If you've already got your parameters in the right form:
                - `model__<arg>` for model parameters
                - `pipeline__<pipe_name>__<arg> for pipeline parameters.

        Args:
            data_sources (Dict[str, Union[DataCatalog, DataSet]]): The DataCatalog or DataSets to use, in DataSource
                form. Should contain at least a 'train' DataSet.
                DataCatalogs will be initialised (`read`) in the beginning of the run. It is recommended to use
                DataCatalog in distributed cases, but for quick experimentation DataSets tend to be more useful.
                Please note that the DataSet will be changed by the Pipeline and Model. If you want to use a
                DataSet, you should make sure you don't modify the original data by reference or by key.
            model (Model): The Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            type_checker_classes (Optional[List[Type[TypeChecker]]]): A list of type checkers. Will be used to
                instantiate TypeCheckerPipe's for input and output datasets.
            input_type_checker (Optional[TypeCheckerPipe]): A type checker for all input data. Will be used to verify
                incoming data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Not required if `type_checker_classes` is set.
            output_type_checker (Optional[TypeCheckerPipe]): A type checker for all output data. Will be used to verify
                outgoing data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Not required if `type_checker_classes` is set.
            serialiser (Optional[Serialiser]): The serialiser to serialise any Python objects (expect the Model).
                Defaults to a joblib serialiser.
            output_folder (Union[Path, str]): The output folder to log artifacts to. Defaults to "outputs".
            additional_files_to_store (Optional[List[Union[str, Path]]]): Extra files to store, such as python files.
                It is possible to select a directory as well, not just individual files.
                Defaults to no extra files (None).
            parameters (Optional[Dict[str, Any]]): Any parameters to log as part of this experiment.
                Defaults to None.
            plots (Optional[List[BasePlotter]]): A list of plots to generate. Defaults to None, resulting in no plots.
                All plots will be written to the outputs folder.
        """
        assert "train" in data_sources, "Must provide a 'train' entry in the data_sources dictionary."
        self.logger = getLogger(__name__)
        if serialiser is None:
            serialiser = JoblibSerialiser()

        if plots is None:
            plots = []

        if additional_files_to_store is None:
            additional_files_to_store = []
        if parameters is None:
            parameters = {}
            self.logger.warning(
                """It is highly recommended to provide the parameters used to initialise your
run here for logging purposes. Consider using the `from_command_line` or
`from_dictionary` initialisation methods"""
            )
        if isinstance(output_folder, str):
            output_folder = Path(output_folder)

        if type_checker_classes is not None:
            # create the type checkers manually
            input_names = pipeline.get_input_datasets_names()
            output_names = model.outputs
            self.input_type_checker = TypeCheckerPipe(
                "_input_type_checker_", input_names, type_checker_classes=type_checker_classes
            )
            self.output_type_checker = TypeCheckerPipe(
                "_input_type_checker_", output_names, type_checker_classes=type_checker_classes
            )
        elif input_type_checker is not None and output_type_checker is not None:
            # use the pre-provided type checkers.
            self.input_type_checker = input_type_checker
            self.output_type_checker = output_type_checker
        else:
            raise ValueError(
                "Either type_checker_classes needs to be set or "
                "input_type_checker AND output_type_checker have to be set."
            )

        self.data_sources = data_sources
        self.model = model
        self.pipeline = pipeline
        self.evaluator = evaluator
        self.experiment_logger = logger
        self.parameters = parameters
        self.serialiser = serialiser
        self.output_folder = output_folder
        self.additional_files_to_store = additional_files_to_store
        self.plots = plots

    def run(self) -> Dict[str, Dict[str, Union[str, float, int, bool]]]:
        """Execute the experiment.

        Returns:
            Dict[str, Dict[str, Union[str, float, int, bool]]]: The performance metrics of this run.
        """
        with self.experiment_logger:
            self.logger.info("Log parameters")
            self.experiment_logger.log_parameters(self.parameters)

            self.logger.info("Load data")
            datasets = {
                name: data_source_set.read() if isinstance(data_source_set, DataCatalog) else data_source_set
                for name, data_source_set in self.data_sources.items()
            }

            self.logger.info("Create input type checker")
            self.input_type_checker.fit(datasets["train"])
            input_checked = {name: self.input_type_checker.transform(ds) for name, ds in datasets.items()}

            self.logger.info("Fit pipeline")
            self.pipeline.fit(input_checked["train"])

            self.logger.info("Transform data")
            transformed = {
                name: self.pipeline.transform(data, is_inference=False) for name, data in input_checked.items()
            }

            self.logger.info("Fit model")
            self.model.fit(transformed["train"])

            self.logger.info("Evaluate model")
            metrics = {name: self.evaluator.evaluate(self.model, data) for name, data in transformed.items()}

            self.logger.info("Create output type checker")
            predicted_train = self.model.transform(transformed["train"])
            self.output_type_checker.fit(predicted_train)

            self.logger.info("Log results.")
            self.experiment_logger.log_run(self, metrics, transformed)
            self.logger.info("Done")
        return metrics

    def copy(self, parameters: Dict[str, Any], seed: int = 1) -> "Experiment":
        """Create a fresh copy of this Experiment.

        The Model & Pipeline will be recreated, so any trained versions will be
        re-initialised.

        Old parameters will be kept, but overwritten if new values are provided.
        This makes it easier to make templates for hyperparameter tuning.

        Args:
            parameters (Dict[str, Any]): New parameters for the Model and Pipeline
                to be re-initialised with.
            seed (int, optional): Training seed. Defaults to 1.

        Returns:
            Experiment: A copy of this experiment, intialised with the new set
                of parameters.
        """
        new_params = self.parameters.copy()
        new_params.update(parameters)
        model_class = self.model.__class__

        model_args = get_args_for_prefix("model__", new_params)
        model = model_class(seed=seed, inputs=self.model.inputs, outputs=self.model.outputs, **model_args)

        pipeline_args = get_args_for_prefix("pipeline__", new_params)
        new_pipeline = self.pipeline.copy(pipeline_args)

        return Experiment(
            # We might want to also deep copy data_sources, evaluator, type checkers, and experiment_logger.
            data_sources=self.data_sources,
            model=model,
            pipeline=new_pipeline,
            evaluator=self.evaluator,
            logger=self.experiment_logger,
            serialiser=self.serialiser,
            output_folder=self.output_folder,
            input_type_checker=self.input_type_checker,
            output_type_checker=self.output_type_checker,
            additional_files_to_store=self.additional_files_to_store,
            parameters=new_params,
        )

    @classmethod
    def from_dictionary(
        cls,
        data_sources: Dict[str, Union[DataCatalog, DataSet]],
        model_class: Type[Model],
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        serialiser: Serialiser,
        model_inputs: List[str],
        model_outputs: List[str],
        parameters: Dict[str, Any],
        type_checker_classes: Optional[List[Type[TypeChecker]]] = None,
        input_type_checker: Optional[TypeCheckerPipe] = None,
        output_type_checker: Optional[TypeCheckerPipe] = None,
        output_folder: Union[Path, str] = "outputs",
        seed: int = 1,
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
    ) -> "Experiment":
        """Creates an Experiment from a dictionary with parameters.

        Args:
            data_sources (Dict[str, Union[DataCatalog, DataSet]]): The DataSets to use, in DataSource form.
                Should contain at least a 'train' DataSet. These will be initialised in the beginning of the run.
            model_class (Type[Model]): The class of the Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            serialiser (Serialiser): The serialiser to serialise any Python objects (expect the Model).
            model_inputs (List[str]): Input dataset names to the model.
            model_outputs (List[str]): Output dataset names to the model.
            parameters (Dict[str, Any]): Any parameters to log as part of this experiment.
                Defaults to None.
            type_checker_classes (Optional[List[Type[TypeChecker]]]): A list of type checkers. Will be used to
                instantiate TypeCheckerPipe's for input and output datasets.
            input_type_checker (Optional[TypeCheckerPipe]): A type checker for all input data. Will be used to verify
                incoming data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Only select datasets required to do a run during inference.
                Not required if `type_checker_classes` is set.
            output_type_checker (Optional[TypeCheckerPipe]): A type checker for all output data. Will be used to verify
                outgoing data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Not required if `type_checker_classes` is set.
            output_folder (Union[Path, str]): The output folder to log artifacts to.
            seed (int): The RNG seed to ensure reproducability.
            additional_files_to_store (Optional[List[Union[str, Path]]]): Extra files to store, such as python files.
                Defaults to no extra files (None).

        Returns:
            Experiment: An Experiment created with the given parameters.
        """
        model_args = get_args_for_prefix("model__", parameters)
        model = model_class(seed=seed, inputs=model_inputs, outputs=model_outputs, **model_args)

        pipeline_args = get_args_for_prefix("pipeline__", parameters)
        pipeline.reinitialise(pipeline_args)

        return Experiment(
            data_sources=data_sources,
            model=model,
            pipeline=pipeline,
            evaluator=evaluator,
            logger=logger,
            serialiser=serialiser,
            output_folder=output_folder,
            type_checker_classes=type_checker_classes,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
            additional_files_to_store=additional_files_to_store,
            parameters=parameters,
        )

    @classmethod
    def from_command_line(
        cls,
        data_sources: Dict[str, Union[DataCatalog, DataSet]],
        model_class: Type[Model],
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        serialiser: Serialiser,
        model_inputs: List[str],
        model_outputs: List[str],
        type_checker_classes: Optional[List[Type[TypeChecker]]] = None,
        input_type_checker: Optional[TypeCheckerPipe] = None,
        output_type_checker: Optional[TypeCheckerPipe] = None,
        output_folder: Union[Path, str] = "outputs",
        seed: int = 1,
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
        fixed_arguments: Optional[Dict[str, Any]] = None,
    ) -> "Experiment":
        """Automatically initialises an Experiment from command line arguments.

        Args:
            data_sources (Dict[str, Union[DataCatalog, DataSet]]): The DataSets to use, in DataSource form.
                Should contain at least a 'train' DataSet. These will be initialised in the beginning of the run.
            model_class (Type[Model]): The class of the Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            serialiser (Serialiser): The serialiser to serialise any Python objects (expect the Model).
            model_inputs (List[str]): Input dataset names to the model.
            model_outputs (List[str]): Output dataset names to the model.
            type_checker_classes (Optional[List[Type[TypeChecker]]]): A list of type checkers. Will be used to
                instantiate TypeCheckerPipe's for input and output datasets.
            input_type_checker (Optional[TypeCheckerPipe]): A type checker for all input data. Will be used to verify
                incoming data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Only select datasets required to do a run during inference.
                Not required if `type_checker_classes` is set.
            output_type_checker (Optional[TypeCheckerPipe]): A type checker for all output data. Will be used to verify
                outgoing data and standardise the order of data. Will be used later to help serialise/deserialise data.
                Not required if `type_checker_classes` is set.
            output_folder (Union[Path, str]): The output folder to log artifacts to.
            seed (int): The RNG seed to ensure reproducability.
            additional_files_to_store (Optional[List[Union[str, Path]]], optional): Extra files to store,
                such as python files. Defaults to no extra files (None).
            fixed_arguments (Optional[Dict[str, Any]]): Arguments that won't be read from command line.
                Useful to pass complex objects that you don't want to optimize. Think of classes, loss functions,
                metrics, etc. Any value in this dict will overwrite values from the command line. These need to be
                presented in the same `model__<arg>` or `pipeline__<step>__<arg>` format as the command line arguments.


        Returns:
            Experiment: An Experiment created with the given parameters from the command line.
        """
        arg_parser = cls._get_cmd_args(model_class, pipeline)
        parsed_args, _ = arg_parser.parse_known_args()

        arguments = parsed_args.__dict__
        if fixed_arguments is not None:
            arguments.update(fixed_arguments)

        return cls.from_dictionary(
            data_sources=data_sources,
            model_class=model_class,
            pipeline=pipeline,
            evaluator=evaluator,
            logger=logger,
            serialiser=serialiser,
            output_folder=output_folder,
            type_checker_classes=type_checker_classes,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
            additional_files_to_store=additional_files_to_store,
            parameters=parsed_args.__dict__,
            seed=seed,
            model_inputs=model_inputs,
            model_outputs=model_outputs,
        )

    @classmethod
    def _get_cmd_args(cls, model_class: Type[Model], pipeline: Pipeline) -> ArgumentParser:
        parser = ArgumentParser()

        model_class.get_parameters(parser)
        add_args_to_parser_for_pipeline(parser, pipeline)
        return parser

    def __str__(self) -> str:
        """Create string representation of this Experiment.

        Returns:
            str: A string representation of this Experiment.
        """
        data_part = "datasets:\n" + "\n\t".join(
            [
                f"""\t{name}:
{ds_source.__str__(indents=2)}"""
                for name, ds_source in self.data_sources.items()
            ]
        )

        pipeline_part = "pipeline:\n" + self.pipeline.__str__(indents=1)
        model_part = f"model:\n\t{str(self.model)}"
        logger_part = f"logger:\n\t{str(self.experiment_logger)}"
        evaluator_part = f"evaluator:\n{self.evaluator.__str__(indents=1)}"

        return "\n\n".join([data_part, pipeline_part, model_part, logger_part, evaluator_part])



"""ExperimentLogger records the results of an experiment.

This allows you to integrate with different reporting tools, like MLflow.
"""
from .experiment_logger import ExperimentLogger
from .local_logger import LocalLogger



"""Provides tools for temporarily switching Python over to a different directory."""
import importlib
import logging
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType
from typing import Generator, List, Optional, Set, Union


@contextmanager
def switch_directory(directory: Union[str, Path]) -> Generator:
    """Allows you to temporarily switch Python over to a different directory.

    This changes the following:
        - The current working directory will change to `directory`

    These changes will be undone once the context manager ends.

    Args:
        directory (Union[str, Path]): The directory to switch to.

    Yields:
        Generator: An empty generator.
    """
    # create backups of the workspace and path.
    target_directory = Path(directory)
    old_workspace = os.getcwd()
    os.chdir(target_directory)
    # Yield so the normal code in the managed context can run.
    yield

    os.chdir(old_workspace)


@contextmanager
def switch_workspace(
    target_workspace: Union[str, Path], extra_files: Optional[List[Union[str, Path]]] = None
) -> Generator:
    """Allows you to temporarily switch Python over to using a different working directory.

    This changes the following:
        - The current working directory will change to `target_workspace`
        - `target_workspace` is added to the system path
        - Any files provided in `extra_files` will be assumed to follow a relative path
            from `target_workspace` to python files that will be imported. Variables
            (incl. classes and functions) defined in those files will be imported into
            the __main__ module, and the files will be added to the path. We do not overwrite variables
            already existing in __main__.

    These changes will be undone once the context manager ends.

    Args:
        target_workspace (Union[str, Path]): The directory to switch to.
        extra_files (Optional[List[Union[str, Path]]]): Any additional files whose contents should
            be added to the current python program. Defaults to no extra files.

    Yields:
        Generator: An empty generator.
    """
    if extra_files is None:
        extra_files = []

    # create backups of the workspace and path.
    target_workspace = Path(target_workspace)
    old_workspace = os.getcwd()
    old_sys_path = sys.path.copy()

    # Add each file to  the path, and import variables specific to that file into globals.
    main_lib = importlib.import_module("__main__")
    main_to_remove = _create_temporary_workspace(target_workspace, extra_files, main_lib)
    # Yield so the normal code in the managed context can run.
    yield

    _reset_workspace(old_workspace, old_sys_path, main_lib, main_to_remove)


def _find_all_py_files(files: List[Union[str, Path]]) -> List[str]:
    """Finds all .py files in the given list of files/directories.

    This will not walk directories and only look at directly provided .py
    files. Directories should be importable through sys.path.

    This is mainly done to be able to import objects defined in the training
    script.

    Args:
        files (List[Union[str, Path]]): A list of files / directories for which
            we want to collect the python files.

    Returns:
        List[str]: The list of python files from the list.
    """
    result = []
    for file in files:
        file_p = Path(file)
        assert file_p.exists(), f"`{file}` not found!"
        if Path(file_p).is_file():
            if str(file).endswith(".py"):
                result.append(str(file))
    return result


def _reset_workspace(
    old_workspace: str, old_sys_path: List[str], main_lib: ModuleType, main_to_remove: Set[str]
) -> None:
    os.chdir(old_workspace)
    sys.path = old_sys_path
    for var in main_to_remove:
        del main_lib.__dict__[var]


def _create_temporary_workspace(
    target_workspace: Path, extra_files: List[Union[str, Path]], main_lib: ModuleType
) -> Set[str]:
    """Perform the actual workspace switching.

    Args:
        target_workspace (Path): The directory to switch to.
        extra_files (List[Union[str ,  Path]]): Any additional files whose contents should
            be added to the current python program. Empty lists are accepted.
        main_lib (ModuleType): The __main__ module.

    Returns:
        Set[str]: The set of variable names that should be removed from __main__ again
            after the context manager quits.
    """
    logger = logging.getLogger(__name__)
    main_to_remove = set()

    # change directory and path
    os.chdir(target_workspace)
    sys.path.append(str(target_workspace))

    extra_files_walked = _find_all_py_files(extra_files)
    for file in extra_files_walked:
        if not file.endswith(".py"):
            logger.warning(f"Not adding {file} to path since it is not a python file!")
            continue

        # add the python file to the path. This makes sure we can import variables from other files.
        relative_file_name = str(os.path.join(".", file))
        sys.path.append(relative_file_name)

        # import the file as a module.
        module_name = file.replace("/", ".").replace(".py", "")
        lib = importlib.import_module(module_name)

        # find all variables defined in this file, avoiding things like numpy/pandas imports.
        sub_vars = {
            name: value
            for name, value in lib.__dict__.items()
            if hasattr(value, "__module__") and value.__module__ == module_name
        }

        # add the variable to the main module. This makes sure we can import variables from the
        # same 'main' file that was used to run the experiment.
        # out of safety reasons we only add a variable if it doesn't exist in 'main' yet.
        for var_name, value in sub_vars.items():
            if var_name not in main_lib.__dict__:
                main_to_remove.add(var_name)
                main_lib.__dict__[var_name] = value
    return main_to_remove



"""
The MLPype base package provides all the core components to run a one stop shop ML experiment.

Most of these are configured
as interfaces, allowing the library to be easily extended.

Core classes to look into are:

- The Experiment class. The main functionality resides in `experiment.Experiment`. \
    The Experiment class will combine the other components of this library and run a standardised ML experiment.
- The DataCatalog. Experiments use DataCatalogs to load data. This allows you to define your data as config, \
    so it doesn't immediately need to be loaded into memory.
- The Pipeline and Pipe classes. These are very similar to the sklearn Pipeline, with the exception that they handle \
    data inputs in dictionary format. This gives much more freedom when transforming and combining datasets in the \
        preprocessing step.
- The Evaluator class. This is a standardised way of evaluating models. The BaseEvaluator uses a dictionary of \
    {name: function} to evaluate a model.

Core interfaces to look into are:

- The Model class. This is the base class for all models. The setup is such that it should allow you to easily \
    integrate other packages into your model. We have already integrated the following packages:
    - sklearn
    - tensorflow / keras
    - pyspark
    - xgboost / lightgbm
- The DataSource class. This is the base class for all data sources. This allows easy extension of inputs to multiple \
    data sources, ranging from flat files to databases, or any other source.
"""
from . import data, deploy, evaluate, experiment, logger, model, pipeline, serialiser, utils



"""Provides a hierarchical Model class, capable of nesting multiple models underneath."""
import types
import typing
from collections import defaultdict
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, Dict, Generic, Iterable, List, Optional, Tuple, Type, TypeVar, Union

from mlpype.base.model.model import Model
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser

T = TypeVar("T", bound=Model)
Data = TypeVar("Data")
logger = getLogger(__name__)


class HierarchicalModel(Model, Generic[T]):
    """The base class for hierarchical models.

    A hierarchical model splits the data into multiple groups.

    Each dataset provided to the model needs to be splittable using the data_splitter function.
    Data is automatically recombined after prediction using the data_merger function.
    """

    HIERARCHICAL_MODEL_FOLDER = "hierarchical_model"
    HIERARCHICAL_MODEL_SUBFOLDER = "models"
    SPLITTER_FILENAME = "data_splitter.pkl"
    MERGER_FILENAME = "data_merger.pkl"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        data_splitter: Callable[[Data], Dict[str, Data]],
        data_merger: Callable[[Dict[str, Data]], Data],
        seed: int = 1,
        model: Optional[Dict[str, T]] = None,
        **model_args: Any,
    ) -> None:
        """The base class for hierarchical models.

        A hierarchical model splits the data into multiple groups.

        Each dataset provided to the model needs to be splittable using the data_splitter function.
        Data is automatically recombined after prediction using the data_merger function.

        Each submodel will be created on the fly during training. You can provide `model` if you
        already have a trained dictionary.

        Below are some examples for how to do this yourself.

        ```python
        # Works
        from mlpype.sklearn.model import LinearRegressionModel
        class HierarchicalLRModel(HierarchicalModel[LinearRegressionModel]):
            pass

        # An alternative to dynamically generate the model, which is easier to export/import
        model_class = HierarchicalModel.class_from_model_class(LinearRegressionModel)

        # Unfortunately, using something like the following will not work due to how Generic types are handled.
        HierarchicalLRModel = HierarchicalModel[LinearRegressionModel]
        ```

        Args:
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            data_splitter (Callable[[Data], Dict[str, Data]]): A function to split incoming Data into
                keyed slices.
            data_merger (Callable[[Dict[str, Data]], Data]): A function to merge keyed Data slices into
                a single Data object.
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
            model (Optional[Dict[str, T]], optional): Optional dictionary of pretrained submodels.
                If not set, an empty list will be used (and the model needs to be trained first).
            **model_args (Any): Keyword arguments that will be passed to each `model` to initialise them
                during training. Besides these, `inputs`, `outputs`, and `seed` will be send by default.
        """
        if model is None:
            model = {}
        self.model = model

        super().__init__(inputs, outputs, seed)

        self.data_splitter = data_splitter
        self.data_merger = data_merger
        self.model_args = model_args

    @classmethod
    def _get_annotated_class(cls) -> Type[T]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def _split_data(self, *data: Data) -> Dict[str, List[Data]]:
        logger.info("Splitting dataset")
        split_data = defaultdict(list)
        for d in data:
            split = self.data_splitter(d)
            for k, v in split.items():
                split_data[k].append(v)

        number_of_different_dataset = {len(v) for v in split_data.values()}
        assert len(number_of_different_dataset) == 1, f"Data is not split correctly, got {number_of_different_dataset}"
        return split_data

    def _merge_data(self, data: Dict[str, Union[Tuple[Data], Data]]) -> List[Data]:
        logger.info("Merging dataset")

        data_list: List[Dict[str, Data]] = []
        for key, d in data.items():
            if len(self.outputs) == 1:
                d = (d,)  # type: ignore
            assert isinstance(d, tuple)

            if len(data_list) == 0:
                data_list = [{} for _ in range(len(d))]

            for i, subset in enumerate(d):
                data_list[i][key] = subset

        return [self.data_merger(subdict) for subdict in data_list]

    def _fit(self, *data: Any) -> None:
        split_data = self._split_data(*data)

        self.model = {}
        logger.info("Fitting model")
        for key, dataset in split_data.items():
            logger.info(f"Fitting for key: {key}")
            model = self._get_annotated_class()(
                inputs=self.inputs,
                outputs=self.outputs,
                seed=self.seed,
                **self.model_args,
            )
            model._fit(*dataset)
            self.model[key] = model

    def _transform(self, *data: Data) -> Union[Iterable[Data], Data]:
        split_data = self._split_data(*data)

        result = {}
        for key, dataset in split_data.items():
            logger.info(f"Transforming for key: {key}")
            if key not in self.model:
                logger.error(f"Model for key {key} not found. Please fit model first!")
            model = self.model[key]
            result[key] = model._transform(*dataset)

        res = self._merge_data(result)
        if len(self.outputs) == 1:
            # If we only return 1 output, the `transform` functions expects no tuple
            return res[0]
        return res

    def _save(self, folder: Path) -> None:
        model_folder = folder / self.HIERARCHICAL_MODEL_FOLDER
        model_folder.mkdir(exist_ok=True)

        joblib_serialiser = JoblibSerialiser()
        joblib_serialiser.serialise(self.data_splitter, model_folder / self.SPLITTER_FILENAME)
        joblib_serialiser.serialise(self.data_merger, model_folder / self.MERGER_FILENAME)

        model_subfolder = model_folder / self.HIERARCHICAL_MODEL_SUBFOLDER
        model_subfolder.mkdir(exist_ok=True)
        for name, model in self.model.items():
            model.save(model_subfolder / name)

    @classmethod
    def _load(
        cls: Type["HierarchicalModel"],
        folder: Path,
        inputs: List[str],
        outputs: List[str],
    ) -> "HierarchicalModel":
        model_subclass = cls._get_annotated_class()
        model_folder = folder / cls.HIERARCHICAL_MODEL_FOLDER
        assert model_folder.is_dir(), f"Folder {model_folder} does not exist"

        joblib_serialiser = JoblibSerialiser()
        data_splitter = joblib_serialiser.deserialise(model_folder / cls.SPLITTER_FILENAME)
        data_merger = joblib_serialiser.deserialise(model_folder / cls.MERGER_FILENAME)

        model_subfolder = model_folder / cls.HIERARCHICAL_MODEL_SUBFOLDER
        assert model_subfolder.is_dir(), f"Folder {model_subfolder} does not exist"

        models = {model_name.name: model_subclass.load(model_name) for model_name in model_subfolder.iterdir()}

        return cls(
            inputs=inputs,
            outputs=outputs,
            data_splitter=data_splitter,
            data_merger=data_merger,
            model=models,
        )

    def set_seed(self) -> None:
        """Sets the seed for all submodels.

        Since no models are created at creation time, this will not work. Instead, set_seed is called
        automatically when the models are created at training time.

        This function will reset each seed.
        """
        for model in self.model.values():
            model.set_seed()

    @classmethod
    def class_from_model_class(
        cls,
        model_class: Type[Model],
    ) -> Type["HierarchicalModel"]:
        """Create a HierarchicalModel wrapper class from a Model class.

        This should support all MLpype compliant models.

        Args:
            model_class (Type[Model]): The class of the MLpype submodel. For example,
                LinearRegressionModel or LogisticRegressionModel from the sklearn subpackage.

        Returns:
            Type[HierarchicalModel]: The created HierarchicalModel.
        """
        new_name = f"Hierarchical{model_class.__name__}"
        try:
            old_docs = cls.__doc__
            assert isinstance(old_docs, str)
        except AttributeError:
            logger.warning(f"Failed to add docstring to {new_name}.")
            old_docs = """The base class for hierarchical models.

    A hierarchical model splits the data into multiple groups.

    Each dataset provided to the model needs to be splittable using the data_splitter function.
    Data is automatically recombined after prediction using the data_merger function.
    """

        new_doc = (
            old_docs
            + f"""
See HierarchicalModel for the original source and docs for subfunctions.

This model uses a fixed class: {model_class.__name__}.
The source module is: {model_class.__module__}
"""
        )

        klass = types.new_class(
            new_name,
            (HierarchicalModel[model_class],),  # type: ignore
            exec_body=lambda ns: ns.update(
                {
                    "__doc__": new_doc,
                    "__module__": HierarchicalModel.__module__,
                }
            ),
        )

        return klass

    @classmethod
    def from_model_class(
        cls,
        model_class: Type[Model],
        inputs: List[str],
        outputs: List[str],
        data_splitter: Callable[[Data], Dict[str, Data]],
        data_merger: Callable[[Dict[str, Data]], Data],
        seed: int = 1,
        **model_args: Any,
    ) -> "HierarchicalModel":
        """Create a HierarchicalModel from a Model class.

        This should support all MLpype compliant models.

        Args:
            model_class (Type[Model]): The class of the MLpype submodel. For example,
                LinearRegressionModel or LogisticRegressionModel from the sklearn subpackage.
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            data_splitter (Callable[[Data], Dict[str, Data]]): A function to split incoming Data into
                keyed slices.
            data_merger (Callable[[Dict[str, Data]], Data]): A function to merge keyed Data slices into
                a single Data object.
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
            **model_args (Any): Keyword arguments that will be passed to each `model` to initialise them
                during training. Besides these, `inputs`, `outputs`, and `seed` will be send by default.

        Returns:
            HierarchicalModel: The created HierarchicalModel.
        """
        new_cls = cls.class_from_model_class(model_class)
        return new_cls(
            inputs=inputs,
            outputs=outputs,
            seed=seed,
            data_splitter=data_splitter,
            data_merger=data_merger,
            model=None,
            **model_args,
        )



"""Evaluates a Model on the given Functions."""
from typing import Callable, Dict, Optional, TypeVar, Union

from mlpype.base.data import DataSet
from mlpype.base.evaluate.base_evaluator import BaseEvaluator
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline

Data = TypeVar("Data")


class Evaluator(BaseEvaluator[Data]):
    """Evaluates a Model on the given Functions.

    This is the default Evaluator, and should fit most models and packages.
    It assumes you provide functions that take in (y_true, y_pred) and return a float, int, string, or boolean value.

    Most sklearn evaluation metrics and those from similar packages should be compliant with this format, and otherwise
    it's not too difficult to adept them to this format. Spark for instance does things differently and has its
    own implementation. For example, in sklearn you can do:

    ```python
    from mlpype.base.evaluate import Evaluator
    from sklearn.metrics import mean_squared_error
    from sklearn.metrics import mean_absolute_error
    from sklearn.metrics import mean_absolute_percentage_error

    evaluator = Evaluator(
        {
            "mse": mean_squared_error,
            "mae": mean_absolute_error,
            "mape": mean_absolute_percentage_error,
        }
    )
    ```

    The advantage of this is that this can easily be applied to each input dataset you use, so evaluations
    can be replicated with ease across datasets, so all metrics are directly available.
    """

    def __init__(
        self,
        functions: Dict[str, Callable[[Data, Data], Union[float, int, str, bool]]],
    ) -> None:
        """Evaluates a Model on the given Functions.

        Args:
            functions (Dict[str, Callable[[Data, Data], Union[float, int, str, bool]]]):
                A dict of metric names and metric functions. We expect each to return
                a float, int, string, or boolean value. We provide the arguments as follows:
                    - first, all labeled data, e.g. y_true.
                    - second, all predicted data, e.g. y_pred.
        """
        self.functions = functions

    def evaluate(
        self, model: Model, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the model does not need to be transformed anymore if pipeline is None.

        Args:
            model (Model): The Model to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """
        if pipeline is not None:
            data = pipeline.transform(data)

        predictions = model.transform(data).get_all(model.outputs)
        output_data = data.get_all(model.outputs)

        return {name: func(*output_data, *predictions) for name, func in self.functions.items()}

    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Evaluator.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Evaluator.
        """
        tabs = "\t" * indents
        return "\n".join([f"{tabs}{name}: {func.__name__}" for name, func in self.functions.items()])



"""Contains tools to create plots for mlpype experiments."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Protocol, Union

from mlpype.base.data.dataset import DataSet


class PlotFunction(Protocol):
    """A function definition for creating plots from data."""

    def __call__(self, path: Path, *data: Any) -> None:
        """Creates a plot from the given data and writes it to the given path.

        Args:
            path (Path): The path to save the plot to.
            *data (Any): Datasets to be used in plotting.
        """


class BasePlotter(ABC):
    """A base class for creating plots."""

    @abstractmethod
    def plot(self, plot_folder: Path, data: DataSet, experiment: Any) -> List[Path]:
        """Creates a plot from the given data and writes it to the given path.

        Args:
            plot_folder (Path): The folder to write the plot to. You still need to set your
                file name. If multiple plots are created, you can use the same folder, but
                make sure to return all file names in the list.
            data (DataSet): The full dataset to plot. This should contain all the data you
                need to make your plots. It contains the last DataSet from the pipeline,
                with the predictions added as "{output_name}{Constants.PREDICTION_POSTFIX}"
            experiment (Any): The experiment object. Useful for extracting
                the trained model.

        Returns:
            List[Path]: The file paths of where the plot(s) are stored.
        """
        raise NotImplementedError


@dataclass
class Plotter(BasePlotter):
    """Creates 1 plot from data."""

    plot_function: PlotFunction
    file_name: Union[Path, str]
    dataset_names: List[str]

    def plot(self, plot_folder: Path, data: DataSet, experiment: Any) -> List[Path]:
        """Selects the correct data from the DataSet, creates a plot, and writes it to `plot_folder / self.file_name`.

        This is a simple wrapper to make a single plot.

        Args:
            plot_folder (Path): The folder to write the plot to. The final plot will be written to
                `plot_folder / self.file_name`.
            data (DataSet): A DataSet containing all the data you need to make your plots.
                In an Experiment, this will contain the last DataSet from the Pipeline with the
                predictions added as "{output_name}{Constants.PREDICTION_POSTFIX}"
            experiment (Any): The experiment object. Not used by default for this
                plotter; it assumens you only use the transformed input data.

        Returns:
            List[Path]: The file paths of where the plot(s) are stored.
        """
        plot_ds = data.get_all(self.dataset_names)
        plot_file = plot_folder / self.file_name
        self.plot_function(plot_file, *plot_ds)
        return [plot_file]



"""The core of the mlpype library: run a standardised ML experiment with the given parameters.

The Experiment defines everything you need to run a standardised ML experiment. It will then do
the following:

- Load data
- Fit the pipeline
- Transform data
- Fit the model
- Evaluate the model
- Log the results and artifacts
"""
from .experiment import Experiment



"""Base class for an Operator that can be used in a Pipeline."""
from abc import ABC, abstractmethod
from typing import Generic, Tuple, TypeVar

Data = TypeVar("Data")


class Operator(ABC, Generic[Data]):
    """An Operator that can been fitted to Data and transform new data.

    This is not meant to be extended, but should work well enough to help
    people identify what their Pipes need as input. sklearn pipe segments
    are supported, like StandardScaler. Custom Operators can be made
    by extending this class.
    """

    def __init__(self) -> None:
        """An Operator that can been fitted to Data and transform new data."""
        super().__init__()

    @abstractmethod
    def fit(self, *data: Data) -> "Operator":
        """Fit the Operator to the given Data.

        Args:
            *data (Data): all Data to be used to fit this pipe.

        Returns:
            Operator: this object.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, *data: Data) -> Tuple[Data, ...]:
        """Transforms the data using this Operator.

        You do not need to return the same amount of Data as you
        used for input.

        Args:
            *data (Data): all Data to be used to transform using this pipe.

        Returns:
            Tuple[Data, ...]: A Tuple of Data elements, the transformed
                Data.
        """
        raise NotImplementedError

    def inverse_transform(self, *data: Data) -> Tuple[Data, ...]:
        """Optionally, you can include an option to inverse transform the data.

        Args:
            *data (Data): all Data to be used to inverse transform using this pipe.

        Returns:
            Tuple[Data, ...]: An Tuple of Data elements, the inverse transformed
                Data.
        """
        return data



"""Provides the Pipeline class.

This class provides a pipeline of operations that can be re-applied to new, similar Data.

This custom pipeline class has been made to account for dictionary inputs, giving you
flexibility in how you build your pipeline and input data.
"""
import logging
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from mlpype.base.data import DataSet
from mlpype.base.pipeline.pipe import Pipe
from mlpype.base.utils.parsing import get_args_for_prefix


class Pipeline:
    """A pipeline of operations that can be re-applied to new, similar Data.

    This custom pipeline class has been made to account for dictionary inputs, giving you
    flexibility in how you build your pipeline and input data. It allows for multiple
    inputs, outputs, and even fit inputs that are only used during fitting, while
    still giving an interface similar to sklearn and other pipeline tools.

    It consists of a list of either Pipe objects or other Pipeline objects. Pipe objects
    are a single step in the pipeline, while Pipelines are an entire pipeline. In the end,
    Pipe's form a thin layer around an Operator to act as an interface between the
    Operator and the DataSet. This allows you to reuse existing pipeline segments from sklearn
    and similar tools.

    By default, we'll log at INFO level which step we're at in fit, transform,
    and inverse_transform.
    """

    def __init__(self, pipes: List[Union[Pipe, "Pipeline"]]) -> None:
        """A pipeline of operations that can be re-applied to new, similar Data.

        Args:
            pipes (List[Union[Pipe, Pipeline]]): A list of either Pipes
                or other Pipelines. These form the steps that should be applied.
        """
        self.pipes = pipes
        self._assert_all_names_different()
        self.logger = logging.getLogger(__name__)

    def _assert_all_names_different(self, names: Optional[set] = None) -> None:
        if names is None:
            names = set()

        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                pipe._assert_all_names_different(names)
            else:
                assert pipe.name not in names, f"{pipe.name} is used multiple times."
                names.add(pipe.name)

    def fit(self, data: DataSet) -> "Pipeline":
        """Fits the entire Pipeline to the given DataSet.

        Args:
            data (DataSet): The DataSet to use in fitting.

        Returns:
            Pipeline: This object.
        """
        for pipe in self.pipes:
            if isinstance(pipe, Pipe):
                self.logger.info(f"\tFitting step: `{pipe.name}`")
            pipe.fit(data)
            data = pipe.transform(data, is_inference=False)
        return self

    def transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Transforms the given data using this Pipeline.

        This Pipeline should be fitted first using fit().

        Args:
            data (DataSet): The DataSet to use in transforming.
            is_inference (bool, optional): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The Transformed Data.
        """
        for pipe in self.pipes:
            if isinstance(pipe, Pipe):
                self.logger.info(f"\tTransforming step: `{pipe.name}`")
            data = pipe.transform(data, is_inference=is_inference)
        return data

    def inverse_transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Inverse transforms the DataSet using this Pipeline.

        Note that this is automatically done in reverse: the inverse steps
        will be applied from back to front.

        Args:
            data (DataSet): The DataSet to use in inverse transforming.
            is_inference (bool, optional): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The inverse transformed DataSet
        """
        for pipe in reversed(self.pipes):
            if isinstance(pipe, Pipe):
                self.logger.info(f"\tInverse transforming step: `{pipe.name}`")
            data = pipe.inverse_transform(data, is_inference=is_inference)
        return data

    def reinitialise(self, args: Dict[str, Any]) -> None:
        """Re-initialises this Pipeline's Pipes using the given dictionary.

        Args:
            args (Dict[str, Any]): Dictionary containing new arguments.
                The keys of the `args` dict should be of the form `<name>__<arg_name>`:
                - name: the name of the Pipe.
                - arg_name: the argument name to use.
                The values will be used as argument parameters.
        """
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                pipe.reinitialise(args)
            else:
                pipe_args = get_args_for_prefix(f"{pipe.name}__", args)
                if len(pipe_args) > 0:
                    pipe.reinitialise(pipe_args)

    def copy(self, args: Optional[Dict[str, Any]] = None) -> "Pipeline":
        """Create a copy of this Pipeline using the given dictionary.

        Args:
            args (Optional[Dict[str, Any]]): Dictionary containing new arguments.
                The keys of the `args` dict should be of the form `<name>__<arg_name>`:
                - name: the name of the Pipe.
                - arg_name: the argument name to use.
                The values will be used as argument parameters.

        Returns:
            Pipeline: A copy of this pipeline.
        """
        result = Pipeline([])
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                extra_pipeline = pipe.copy(args)
                result += extra_pipeline
            else:
                pipe_args = pipe.args.copy()
                if args is not None:
                    pipe_args.update(get_args_for_prefix(f"{pipe.name}__", args))
                new_pipe = pipe.copy(pipe_args)
                result += new_pipe
        return result

    def __iter__(self) -> "Pipeline":
        """Prepares this object for iteration using next().

        Returns:
            Pipeline: this object.
        """
        self.__iter_n = 0
        return self

    def __next__(self) -> Union["Pipeline", Pipe]:
        """Gets the next item in this Pipeline.

        Raises:
            StopIteration: If this object has already been iterated.

        Returns:
            Union[Pipeline, Pipe]: The next Pipeline/Pipe.
        """
        if self.__iter_n >= len(self.pipes):
            raise StopIteration
        else:
            pos = self.__iter_n
            self.__iter_n += 1
            return self[pos]

    def __getitem__(self, pos: Union[int, str, slice]) -> Union["Pipeline", Pipe]:
        """Gets the Pipe/Pipeline at the given position.

        Slices and ints will be passed directly to the underlying list. Strings
        will be searched for in the names of the Pipes in this and any nested Pipelines
        in this object.

        Args:
            pos (Union[int, str, slice]): the position to retrieve.

        Raises:
            KeyError: if `pos` does not exist in this object.
                int: `pos` is out of bounds.
                str: the given name does not match any Pipe in this object.
                    Any Pipelines in this object are also searched. This makes it
                    important to use different names for Pipes!

        Returns:
            Union[Pipeline, Pipe]: The Pipeline/Pipe matching the given position.
        """
        value = self._get(pos)
        if value is None:
            raise KeyError(f"{pos} is not found in this pipeline")
        return value

    def _get(self, pos: Union[int, str, slice]) -> Union["Pipeline", Pipe, None]:
        if isinstance(pos, slice):
            return Pipeline(self.pipes[pos])
        if isinstance(pos, int):
            return self.pipes[pos]
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                result = pipe._get(pos)
                if result is not None:
                    return result
            elif pipe.name == pos:
                return pipe
        return None

    def __add__(self, obj: Union["Pipeline", Pipe]) -> "Pipeline":
        """Append a Pipe or Pipeline object and return a new Pipeline.

        Args:
            obj (Union[Pipeline, Pipe]): The object to be appended. Can be a Union[Pipeline, Pipe].

        Raises:
            TypeError: If `obj` is not a Pipe or Pipeline.

        Returns:
            Pipeline: A Pipeline with the two sets of steps combined. The steps from this
                object will always be used first.
        """
        if isinstance(obj, Pipeline):
            return Pipeline(self.pipes + obj.pipes)
        elif isinstance(obj, Pipe):
            return Pipeline(self.pipes + [obj])
        else:
            raise TypeError("Pipeline can only use + with a Pipe or Pipeline")

    def insert(self, pos: int, pipe: Pipe) -> None:
        """Insert a pipe in the given integer position.

        We check if the pipe name doesn't exist yet.

        Args:
            pos (int): The position to put the new pipe
            pipe (Pipe): The pipe to insert
        """
        self._assert_all_names_different(set(pipe.name))
        self.pipes.insert(pos, pipe)

    def __len__(self) -> int:
        """Return the length of this Pipeline.

        If it contains any sub-pipelines, those will be counted in full as well
        (and thus not as a single step, unless they only contain 1 step).

        Returns:
            int: The lenght of this Pipeline.
        """
        result = 0
        for p in self.pipes:
            if isinstance(p, Pipeline):
                result += len(p)
            else:
                result += 1
        return result

    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Pipeline.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Pipeline.
        """
        tabs = "\t" * indents
        output = [tabs + "Pipeline:"]
        indents += 1
        for pipe in self.pipes:
            if isinstance(pipe, Pipe):
                output.append(tabs + str(pipe))
            else:
                output.append(pipe.__str__(indents))
        return "\n".join(output)

    def get_input_datasets_names(self, add_fit_inputs: bool = False) -> List[str]:
        """Returns all datasets that are considered inputs.

        Args:
            add_fit_inputs (bool, optional): Flag indicating if fit inputs should be added. Defaults to False.

        Returns:
            List[str]: All input dataset names that used in this pipeline.
        """
        inputs, _ = self._get_input_datasets_names(add_fit_inputs=add_fit_inputs)
        result = list(inputs)
        result.sort()
        return result

    def _get_input_datasets_names(
        self, add_fit_inputs: bool, seen_names: Optional[Set[str]] = None
    ) -> Tuple[Set[str], Set[str]]:
        result: Set[str] = set()
        if seen_names is None:
            seen_names = set()

        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                pl_result, pl_seen = pipe._get_input_datasets_names(add_fit_inputs, seen_names)
                result = result | pl_result
                seen_names = seen_names | pl_seen
            else:
                input_set = set(pipe.inputs)
                output_set = set(pipe.outputs)
                if add_fit_inputs:
                    input_set = input_set | set(pipe.fit_inputs)
                # all inputs that are not already part of the currently seen set of names
                new_datasets = input_set - seen_names

                result = result | new_datasets
                seen_names = seen_names | input_set | output_set
        return result, seen_names



"""An interface for logging experiments."""
import json
import logging
import os
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

from mlpype.base.constants import Constants
from mlpype.base.data import DataSet
from mlpype.base.model.model import Model
from mlpype.base.serialiser import JoblibSerialiser, Serialiser

if TYPE_CHECKING:
    from mlpype.base.experiment import Experiment


class ExperimentLogger(ABC):
    """An interface for logging experiments."""

    def __init__(self) -> None:
        """An interface for logging experiments."""
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def __enter__(self) -> None:
        """Start the experiment."""
        return

    def log_run(
        self,
        exp: "Experiment",
        metrics: Dict[str, Dict[str, Union[str, float, int, str, bool]]],
        transformed: Dict[str, DataSet],
    ) -> None:
        """Log the results of a run.

        This method logs:
            - The metrics for each dataset
            - The pipeline, serialiser, and input/output type checkers
            - The model
            - The plots
            - The requirements.txt

        Args:
            exp (Experiment): The experiment to log
            metrics (Dict[str, Dict[str, Union[str, float, int, str, bool]]]): A dictionary of dicts:
                per dataset: metric names and values
            transformed (Dict[str, DataSet]): A dictionary of transformed datasets (after applying the pipeline)
        """
        for dataset_name, metric_set in metrics.items():
            self.logger.info(f"Log results: metrics for {dataset_name}")
            self.log_metrics(dataset_name, metric_set)

        self._create_output_folders(exp.output_folder)

        # Log plots, requirements.txt, and extra files
        self.logger.info("Log plots, requirements.txt, and extra files")
        self._log_plots(exp, transformed)
        self._log_requirements(exp.output_folder)
        self._log_extra_files(exp.output_folder, exp.additional_files_to_store)

        # log serialiser using a JoblibSerialiser
        self.logger.info("Log results: pipeline, serialiser, input/output type checkers")
        jl_serialiser = JoblibSerialiser()
        self.log_artifact(exp.output_folder / Constants.SERIALISER_FILE, jl_serialiser, object=exp.serialiser)
        self.log_artifact(exp.output_folder / Constants.PIPELINE_FILE, exp.serialiser, object=exp.pipeline)
        self.log_artifact(
            exp.output_folder / Constants.INPUT_TYPE_CHECKER_FILE, exp.serialiser, object=exp.input_type_checker
        )
        self.log_artifact(
            exp.output_folder / Constants.OUTPUT_TYPE_CHECKER_FILE, exp.serialiser, object=exp.output_type_checker
        )

        # Finally, log the model. Do this so mlflow can actually be useful...
        self.logger.info("Log model")
        self.log_model(exp.model, exp.output_folder / Constants.MODEL_FOLDER)

    def _log_requirements(self, output_folder: Path) -> None:
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        version_info = {
            "python_version": python_version,
            "major": sys.version_info.major,
            "minor": sys.version_info.minor,
            "micro": sys.version_info.micro,
        }
        version_file = output_folder / Constants.PYTHON_VERSION_FILE
        with open(version_file, "w") as f:
            json.dump(version_info, f)
        self.log_file(version_file)

        reqs = subprocess.check_output([sys.executable, "-m", "pip", "freeze"]).decode()
        requirements_file = output_folder / Constants.REQUIREMENTS_FILE
        with open(requirements_file, "w") as f:
            f.write(reqs)
        self.log_file(requirements_file)

    def _log_extra_files(self, output_folder: Path, additional_files_to_store: List[Union[Path, str]]) -> None:
        """Logs the extra files for an experiment, as specified in the constructor.

        This also supports saving files outside of the current working directory.
        These files are saved under the root name of the file / folder you select.
        """
        paths_to_log = []
        self.logger.info("Log extra files")
        for extra_file in additional_files_to_store:
            try:
                # If the file to log is in the current work directory, use that to find
                # the relative path to the CWD. This is to help imports.
                source_path = Path(extra_file).relative_to(os.getcwd())
                logged_path = str(source_path)
            except ValueError:
                # The file is not on our current working directory. It is assumed you
                # added the file / folder to the system path in an alternative way.
                # It will be added as such.
                source_path = Path(extra_file).absolute()
                logged_path = Path(extra_file).name
            target_path = output_folder / logged_path
            self.log_local_file(source_path, target_path)
            paths_to_log.append(logged_path)

        self.logger.info("Log `extra files`-file")
        extra_files_file = output_folder / Constants.EXTRA_FILES
        with open(extra_files_file, "w") as f:
            data = {"paths": paths_to_log}
            json.dump(data, f)
        # Make sure the file path is closed before logging anything, to make sure all writes have flushed.
        self.log_file(extra_files_file)

    def _log_plots(self, exp: "Experiment", datasets: Dict[str, DataSet]) -> None:
        predicted = {name: exp.model.transform(data) for name, data in datasets.items()}

        full_datasets = {}
        for dataset_name, dataset in predicted.items():
            assert dataset_name in datasets
            transformed_set = datasets[dataset_name]
            for data_name, data in dataset.items():
                new_name = f"{data_name}{Constants.PREDICTION_SUFFIX}"
                assert new_name not in transformed_set
                transformed_set[new_name] = data

            full_datasets[dataset_name] = transformed_set

        for plotter in exp.plots:
            for name, data in full_datasets.items():
                plot_folder = exp.output_folder / Constants.PLOT_FOLDER / name
                plot_folder.mkdir(parents=True, exist_ok=True)
                plot_files = plotter.plot(plot_folder, data, exp)

                for plot_file in plot_files:
                    self.log_file(plot_file)

    def _create_output_folders(self, output_folder: Path) -> None:
        if output_folder.exists():
            self.logger.warning("Output folder already exists. This may cause conflicts.")
        output_folder.mkdir(exist_ok=True, parents=True)
        (output_folder / Constants.MODEL_FOLDER).mkdir(exist_ok=True)

    @abstractmethod
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Stop the experiment."""
        return

    def log_metrics(self, dataset_name: str, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Logs the metrics for a given dataset.

        Metrics will be stored by default as `<dataset_name>_<metric_name>`.

        Args:
            dataset_name (str): The name of the dataset. Will be prepended to any metric name.
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        metrics = {f"{dataset_name}_{name}": metric for name, metric in metrics.items()}
        self._log_metrics(metrics)

    @abstractmethod
    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Perform the actual logging of metrics.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        raise NotImplementedError

    @abstractmethod
    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Logs the parameters for a given run.

        These will by default be passed on with prefixes such as `model__`.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        raise NotImplementedError

    def log_model(self, model: Model, folder: Union[str, Path]) -> None:
        """Logs a Model for a given experiment.

        This function will write the model to the given location as well.

        Args:
            model (Model): The Model to be logged.
            folder (Union[str, Path]): The file to log the Model to.
        """
        model.save(folder)
        self.log_file(folder)

    def log_artifact(self, file: Union[str, Path], serialiser: Serialiser, object: Any) -> None:
        """Logs an artifact/file as part of an experiment.

        Args:
            file (Union[str, Path]): The file to serialise to and subsequently log.
            serialiser (Serialiser): Serialiser used to serialise the given object.
            object (Any): The object to log. This object will be serialised to the given file
                using the serialiser and logged using `log_file`
        """
        serialiser.serialise(object, file)
        self.log_file(file)

    def log_local_file(self, file: Union[str, Path], output_target: Union[str, Path]) -> None:
        """Logs a given local file / directory as part of an experiment.

        First the file / directory is copied to the output directory, then
        log_file is called to make it part of the experiment.

        Args:
            file (Union[str, Path]): The file / directory to log.
            output_target (Union[str, Path]): The target location of the file / directory.
        """
        file = Path(file)
        output_target = Path(output_target)
        output_target.parent.mkdir(exist_ok=True, parents=True)

        if file.is_file():
            shutil.copy(file, output_target)
        else:
            if output_target.exists():
                shutil.rmtree(output_target)
            shutil.copytree(file, output_target)

        self.log_file(output_target)

    @abstractmethod
    def log_file(self, file: Union[str, Path]) -> None:
        """Logs a given file as part of an experiment.

        Args:
            file (Union[str, Path]): The file to log.
        """
        raise NotImplementedError



"""Provides tools for deploying mlpype models.

This provides 2 main tools:

- The Inferencer: Provides a standard way of inferencing with mlpype models. Can use the output of \
    an experiment to load a model back into memory. The easiest way to use this is to use `Inferencer.from_experiment`.
- The wheel: Provides a standard way of turning an mlpype experiment into a wheel file, which can be installed.
"""
from . import wheel
from .inference import Inferencer



"""Provides tools for parsing arguments from docstrings and type hints into command line hints."""
import inspect
import re
import typing
import warnings
from argparse import ArgumentParser
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union

from docstring_parser import parse

from mlpype.base.pipeline.pipe import Pipe
from mlpype.base.pipeline.pipeline import Pipeline


def add_args_to_parser_for_pipeline(parser: ArgumentParser, pipeline: Pipeline) -> None:
    """Adds arguments for a full pipeline re-initialisation to the given ArgumentParser.

    This will add arguments for every Pipe in the pipeline.

    Arguments will be drawn in order from:
        - type hints
        - documentation.

    Documentation hints are not guaranteed to work. For the most reliable results, use type hints.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        pipeline (Pipeline): The pipeline for which re-initialisation arguments should be added.
    """
    for pipe in pipeline:
        if isinstance(pipe, Pipe):
            add_args_to_parser_for_class(parser, pipe.operator_class, f"pipeline__{pipe.name}", [])
        else:
            add_args_to_parser_for_pipeline(parser, pipe)


def add_args_to_parser_for_class(
    parser: ArgumentParser,
    class_: Type,
    prefix: str,
    excluded_superclasses: List[Type],
    excluded_args: Optional[List[str]] = None,
) -> None:
    """Adds arguments for the constructor of a class to the given parser.

    We try to dive deeper in case of variable kw-args, but in general we recommend not
    using these.

    Arguments will be drawn in order from:
        - type hints
        - documentation.

    Documentation hints are not guaranteed to work. For the most reliable results, use type hints.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        class_ (Type): The class who's constructor's arguments we want to add.
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        excluded_superclasses (List[Type]): Superclasses of the class_ that should be ignored in case
            there are kw-args style arguments to the constructor.
        excluded_args (Optional[List[str]], optional): argument names to never include.
            By default 'self' and 'cls' are ignored.
    """
    init_func = class_.__init__
    class_docstring_args = _parse_docs_to_type_args(class_)
    if len(class_docstring_args) == 0:
        class_docstring_args = _parse_docs_to_type_args(init_func)

    add_args_to_parser_for_function(parser, init_func, prefix, excluded_args, class_docstring_args=class_docstring_args)

    signature = inspect.signature(init_func)
    for _, param in signature.parameters.items():
        # we only need to check for superclass arguments if there are var-args.
        # we ignore positional var-args.
        if param.kind == param.VAR_KEYWORD:
            for superclass in class_.__bases__:
                if superclass not in excluded_superclasses:
                    add_args_to_parser_for_class(parser, superclass, prefix, excluded_superclasses, excluded_args)


def add_args_to_parser_for_function(
    parser: ArgumentParser,
    function: Callable,
    prefix: str,
    excluded: Optional[List[str]] = None,
    class_docstring_args: Optional[Dict[str, Union[Type, None]]] = None,
) -> None:
    """Add arguments for a given function to the given parser.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        function (Callable): The function who's arguments we want to add.
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        excluded (Optional[List[str]]): argument names to never include. By default 'self' and 'cls' are ignored.
        class_docstring_args (Optional[Dict[str, Union[Type, None]]]): docstring arguments obtained from the class.
            Useful for init functions, since those can be documented in multiple places.
    """
    args = inspect.signature(function)
    docstring_args = _parse_docs_to_type_args(function)

    if class_docstring_args is not None:
        docstring_args.update(**class_docstring_args)

    for name, parameter in args.parameters.items():
        class_ = parameter.annotation
        if class_ == inspect._empty and name in docstring_args:
            class_ = docstring_args[name]
        required = parameter.default == inspect._empty
        add_argument(parser, name, prefix, class_, required, parameter.default, excluded)


def _parse_docs_to_type_args(
    func: Callable, extra_mappings: Optional[Dict[str, type]] = None, include_none_args: bool = False
) -> Dict[str, Union[type, None]]:
    """Parses the docstring of a function into an arg-type dictionary.

    We only parse the following types, whose first letter can be capitalized.
        - bool
        - str
        - int
        - float

    This is all done on a best-effort basis. We do not guarantee this will work for every
    function in every library. For the best results, make sure you use proper type hinting
    instead.

    Args:
        func (Callable): The function whose docstring needs to be read.
        extra_mappings (Optional[Dict[str, type]], optional): Optional extra regex -> type mappings.
            Defaults to no extra mappings.
        include_none_args (bool, optional): Whether to include arguments in the output
            with no good type mappings. Defaults to False.

    Returns:
        Dict[str, Union[type, None]]: A dictionary mapping argument names to types.
    """
    signature = inspect.signature(func)
    if extra_mappings is None:
        extra_mappings = {}

    result: Dict[str, Union[type, None]] = {}
    doc = inspect.getdoc(func)
    if doc is None:
        return result

    parsed_doc = parse(doc)
    for param in parsed_doc.params:
        if param.arg_name in signature.parameters:
            class_ = _parse_type_name(param.type_name, extra_mappings)
            if include_none_args or class_ is not None:
                result[param.arg_name] = class_

    return result


def _parse_type_name(s: Optional[str], extra_mappings: Optional[Dict[str, type]] = None) -> Union[type, None]:
    if extra_mappings is None:
        extra_mappings = {}
    if s is None:
        return None

    regexes = {
        r"(^|\W)[fF]loat($|\W)": float,
        r"(^|\W)[sS]tr($|\W)": str,
        r"(^|\W)[bB]ool($|\W)": bool,
        r"(^|\W)[iI]nt($|\W)": int,
    }
    regexes.update(**extra_mappings)

    for regex, class_ in regexes.items():
        if re.search(regex, s) is not None:
            return class_
    return None


def add_argument(
    parser: ArgumentParser,
    name: str,
    prefix: str,
    class_: Union[type, None],
    is_required: bool,
    default: Any,
    excluded: Optional[List[str]] = None,
) -> None:
    """Add an argument to the given parser.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        name (str): The name of the argument
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        class_ (Union[type, None]): The class of the argument. Currently we support:
            - str, float, int, bool
            - lists / tuples / iterables of the above 4 arguments.
        is_required (bool): Whether the argument is required or optional.
        default (Any): the default value if is_required is False.
        excluded (Optional[List[str]]): argument names to never include. By default 'self' and 'cls' are ignored.
    """
    if excluded is None:
        excluded = []
    excluded = excluded + ["self", "cls"]

    if not is_required:
        assert default != inspect._empty, "A default value must be provided if the argument is not required"
    else:
        default = None

    arg_name = f"--{prefix}__{name}"
    if name in excluded or class_ is None:
        return
    if class_ in [str, float, int, bool]:
        parser.add_argument(arg_name, type=_get_conversion_function(class_), required=is_required, default=default)
    elif typing.get_origin(class_) in [
        list,
        tuple,
        typing.get_origin(Iterable),
        typing.get_origin(List),
        typing.get_origin(Tuple),
    ]:
        subtype = typing.get_args(class_)[0]
        parser.add_argument(
            arg_name, type=_get_conversion_function(subtype), nargs="+", required=is_required, default=default
        )
    else:
        warnings.warn(f"Currently the class {str(class_)} is not supported for automatic command line importing.")


def _get_conversion_function(class_: type) -> Callable:
    if class_ == bool:
        return _convert_bool
    return class_


def _convert_bool(s: str) -> bool:
    return s.lower() in ["true", "1"]



"""Provides a base class for evaluators."""
from abc import ABC, abstractmethod
from typing import Dict, Generic, Optional, TypeVar, Union

from mlpype.base.data import DataSet
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline

Data = TypeVar("Data")


class BaseEvaluator(Generic[Data], ABC):
    """Evaluates a Model on the given data."""

    def __init__(
        self,
    ) -> None:
        """Evaluates a Model on the given data.

        This is the base class for all evaluators. Most packages however can use the
        Evaluator class directly, to provide an easier interface. That one allows you
        to provide a dictionary of functions that you can then evaluate on your data.
        """

    @abstractmethod
    def evaluate(
        self, model: Model, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the data does not need to be transformed anymore if pipeline is None.

        Args:
            model (Model): The Model to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """

    @abstractmethod
    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Evaluator.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Evaluator.
        """



"""Provides the base classes for pipelines in MLpype."""
from .operator import Operator
from .pipe import Pipe
from .pipeline import Pipeline
from .type_checker import DataModel, DataSetModel, TypeChecker, TypeCheckerPipe



"""Provides Serialisers for mlpype.

The JoblibSerialiser is the default serialiser for mlpype. It should work for most use cases.

Packages like Tensorflow and Keras should use their own Serialisers.
"""
from .joblib_serialiser import JoblibSerialiser
from .serialiser import Serialiser



"""This module contains the base data types for mlpype.

We include:

- DataSource: An interface for reading data. The main way to define used datasets in MLpype.
- DataSink: An interface for writing data.
- DataCatalog: A container for DataSources. This allows you to keep track of all your data sources in one place, and \
    load them all in one go.
- DataSet: A collection of read data. This is used internally to pass your data between pipelines and models.

"""
from .data_catalog import DataCatalog
from .data_sink import DataSink
from .data_source import DataSource
from .dataset import DataSet



"""Provides the core Model class, which is the base class for all mlpype-compliant models."""
import json
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import Generic, List, Tuple, TypeVar, Union

from dill import dump, load

from mlpype.base.constants import Constants
from mlpype.base.data import DataSet

Data = TypeVar("Data")


class Model(ABC, Generic[Data]):
    """An abstraction of a ML model.

    This class is the core for any Models to integrate with the mlpype framework.
    This is the main abstraction layer between most other packages like sklearn and
    mlpype. It allows you to train models and make predictions using the same interface,
    so that you can easily switch between packages and thus models.

    Extend it by implementing the abstract methods:

    - `_fit`: Fit the model to the given data.
    - `_transform`: Transform the given data using the model to make predictions.
    - `_save`: Save the model to the given folder.
    - `_load`: Load the model from the given folder.
    - `set_seed`: Set the RNG seed.

    """

    def __init__(self, inputs: List[str], outputs: List[str], seed: int = 1) -> None:
        """An abstraction of a ML model.

        This should provide the basic interface for fitting, inference, and serialisation.

        Args:
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            seed (int, optional): The RNG seed to ensure reproducability. Defaults to 1.
        """
        super().__init__()
        self.seed = seed
        self.inputs = inputs
        self.outputs = outputs
        self.set_seed()

    @classmethod
    def get_parameters(cls, parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """

    @abstractmethod
    def set_seed(self) -> None:
        """Sets the RNG seed.

        Use this to ensure reproducability.
        """
        raise NotImplementedError

    def save(self, folder: Union[str, Path]) -> None:
        """Stores this model to the given folder.

        This function stores the common inputs and outputs list to the given folder,
        and makes sure it's created. It also stores the model class in the given folder.
        It will call _save to allow further models to specify how they are stored.

        `_save` will be called with the same folder argument as this function.

        Args:
            folder (Union[str, Path]): The folder to store the Model in.
        """
        if isinstance(folder, str):
            folder = Path(folder)

        folder.mkdir(exist_ok=True, parents=True)
        lists_to_save = {
            "inputs": self.inputs,
            "outputs": self.outputs,
        }
        with open(folder / Constants.MODEL_PARAM_FILE, "w") as f:
            json.dump(lists_to_save, f)

        with open(folder / Constants.MODEL_CLASS_FILE, "wb") as f:
            dump(self.__class__, f)
        self._save(folder)

    @abstractmethod
    def _save(self, folder: Path) -> None:
        """Stores this model to the given folder.

        Specifically intended to store the real model artifact, like the sklearn
        model object or keras model.

        Args:
            folder (Path): The folder to store the Model in.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, folder: Union[str, Path]) -> "Model":
        """Loads a model from a folder into this Model.

        This function first loads the common inputs and outputs list from the given folder,
        then loads the model class in the given folder. It will call _load on this class to allow
        models to specify how they are loaded.

        This is inteded to be called like this:

        ```python
        model = Model.load(folder)
        ```

        Args:
            folder (Union[str, Path]): The folder to load the model from.

        Returns:
            Model: The loaded model. It's type is determined by the model class, which should
                be specified in the folder.
        """
        if isinstance(folder, str):
            folder = Path(folder)

        with open(folder / Constants.MODEL_PARAM_FILE) as f:
            lists = json.load(f)

        with open(folder / Constants.MODEL_CLASS_FILE, "rb") as f:
            model_class = load(f)
        return model_class._load(folder, lists["inputs"], lists["outputs"])

    @classmethod
    @abstractmethod
    def _load(cls, folder: Path, inputs: List[str], outputs: List[str]) -> "Model":
        """Loads a model from file into this Model.

        Generally this focusses only on loading the actual model artifact, like the sklearn
        model object or keras model.

        Args:
            folder (Path): The folder to load the model from.
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.

        Returns:
            Model: self
        """
        raise NotImplementedError

    def fit(self, data: DataSet) -> "Model":
        """Fits the Model to the given DataSet.

        The DataSet should contain all inputs and outputs. This calls the `_fit` function
        to actually fit the model, which handles the actual implementation. On the other hand,
        this function handles grabbing the correct inputs and outputs from the DataSet. This allows
        the implementations of `_fit` to just use raw data, like 2 numpy arrays for X and y.

        Args:
            data (DataSet): The DataSet to fit this Model on.

        Returns:
            Model: This model.
        """
        self._fit(*data.get_all(self.inputs), *data.get_all(self.outputs))
        return self

    @abstractmethod
    def _fit(self, *data: Data) -> None:
        """Fits the Model to the given data.

        This function can directly use the raw data to fit the model. For instance, in the
        case of a sklearn model, it can directly use the raw data to fit the model:

        ```python
        def _fit(self, X, y):
            # Assume self.model is a sklearn model
            return self.model.fit(X, y)
        ```

        This is compatible with the MLPype API.

        Args:
            *data (Data): The data to fit using this Model.
        """
        raise NotImplementedError

    def transform(self, data: DataSet) -> DataSet:
        """Applies the Model to the given DataSet.

        The DataSet should contain all inputs. This calls the `_transform` function
        to actually apply the model, which handles the actual implementation. On the other hand,
        this function handles grabbing the correct inputs from the DataSet. This allows
        the implementations of `_transform` to just use raw data. It also makes sure
        that the output is a DataSet by combining it with the `outputs`.

        Args:
            data (DataSet): The DataSet to transform using this Model.

        Returns:
            DataSet: The outputs as a Dataset.
        """
        result = self._transform(*data.get_all(self.inputs))
        if len(self.outputs) == 1:
            result = [result]  # type: ignore

        # type check handled by above check.
        return DataSet.from_dict({name: data for name, data in zip(self.outputs, result)})  # type: ignore

    @abstractmethod
    def _transform(self, *data: Data) -> Union[Tuple[Data], Data]:
        """Applies the Model to the given data.

        This function can use raw data, like 2 numpy arrays for X and y. In the case of
        a sklearn model, you could define `_transform` like:
        ```python
        def _transform(self, X, y):
            # Assume self.model is a sklearn model
            return self.model.fit(X, y)
        ```

        This will be compatible with the MLPype API.

        Args:
            *data (Data): The DataSet to transform using this Model.

        Returns:
            Union[Tuple[Data], Data]: Either a tuple of Data or a single Data object.
                For instance, this can directly return the predictions of a sklearn model.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        """Create string representation of this Model.

        Returns:
            str: A string representation of this Model.
        """
        return f"{type(self).__name__}: {self.inputs} -> {self.outputs}"



"""Provides tools for evaluating mlpype models.

Evaluators are used to evaluate models in mlpype. The Evaluator class itself can be provided
a dictionary of {name: function} to evaluate a model. Evaluators are run against
all datasets of an experiment.
"""
from .base_evaluator import BaseEvaluator
from .evaluator import Evaluator
from .plot import BasePlotter, Plotter



"""A basic logger which only saves local copies of the data and prints results."""
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from .experiment_logger import ExperimentLogger


class LocalLogger(ExperimentLogger):
    """A basic logger which only saves local copies of the data and prints results."""

    def __enter__(self) -> None:
        """Start the experiment."""

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """Stop the experiment."""

    def __init__(self, min_whitespace: int = 30) -> None:
        """A basic logger which only saves local copies of the data and prints results."""
        super().__init__()
        self.min_whitespace = min_whitespace

    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Prints the metrics to the console.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        for name, value in metrics.items():
            print(f"{name.ljust(self.min_whitespace)}: {value}")

    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Prints the parameters to the console.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        for name, value in parameters.items():
            print(f"{name.ljust(self.min_whitespace)}: {value}")

    def log_file(self, file: Union[str, Path]) -> None:
        """This method does no additional logging.

        Args:
            file (Union[str, Path]): The file to log.
        """



"""Provides a standard way of inferencing with mlpype models."""
import json
from pathlib import Path
from typing import Optional, Type, Union

from mlpype.base.constants import Constants
from mlpype.base.data import DataCatalog, DataSet
from mlpype.base.experiment.experiment import Experiment
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline
from mlpype.base.pipeline.type_checker import TypeCheckerPipe
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser
from mlpype.base.serialiser.serialiser import Serialiser
from mlpype.base.utils.workspace import switch_workspace


class Inferencer:
    """Provides a standard way of inferencing with mlpype models.

    This is the standard way of using the model in production. You can create
    an Inferencer by:
    - Loading one from the output folder from a mlpype experiment. (recommended)
    - Loading one from an experiment.
    - Creating one directly from a model, pipeline, and type checkers.

    It then provides a standardised `predict` method for inference. You can provide a
    DataSet or DataCatalog, and it will return a DataSet with the predictions from
    the model, with transformed data from the pipeline if requested.
    """

    def __init__(
        self,
        model: Model,
        pipeline: Pipeline,
        input_type_checker: TypeCheckerPipe,
        output_type_checker: TypeCheckerPipe,
    ):
        """Provides a standard way of inferencing with mlpype models.

        Args:
            model (Model): The Model to use in inference.
            pipeline (Pipeline): The Pipeline to use in inference.
            input_type_checker (TypeCheckerPipe): The type checker used to determine incoming types.
            output_type_checker (TypeCheckerPipe): The type checker used to determine outgoing types.
        """
        self.model = model
        self.pipeline = pipeline
        self.input_type_checker = input_type_checker
        self.output_type_checker = output_type_checker

    def predict(self, data: Union[DataSet, DataCatalog], return_transformed_data: bool = False) -> DataSet:
        """Predicts using the given data using the Pipeline and Model.

        Args:
            data (Union[DataSet, DataCatalog]): The data to predict for.
            return_transformed_data (bool): Flag indicating if transformed data
                also needs to be returned. If set to True, we still only return one
                dataset, but we'll add the data after the pipeline has run to it.

        Returns:
            DataSet: The predictions from Model, with transformed data from the Pipeline if
                requested.
        """
        # TODO: inverse transformation after prediction
        if isinstance(data, DataCatalog):
            data = data.read()
        checked_input = self.input_type_checker.transform(data)
        transformed = self.pipeline.transform(checked_input, is_inference=True)
        predicted = self.model.transform(transformed)
        checked_output = self.output_type_checker.transform(predicted)

        if return_transformed_data:
            transformed.set_all(checked_output.keys(), checked_output.values())
            return transformed
        return checked_output

    @classmethod
    def from_folder(cls: Type["Inferencer"], folder: Path, serialiser: Optional[Serialiser] = None) -> "Inferencer":
        """Loads a Inferencer from the results of an Experiment.

        This is the recommended way of loading an Inferencer from an experiment to minimise
        the chances of loading issues. This function does the following:

        - Import extra files as specified when training the model from the given folder. This
            will change the current working directory to the given folder temporarily, and will
            load in the content of the extra python files you specified in `extra_files`.
        - Load the model from the given folder.
        - Load the pipeline from the given folder.
        - Load the type checkers from the given folder.
        - Return an Inferencer using the loaded model, pipeline, and type checkers.

        We use the absolute version of the path to try and prevent loading issues.

        Args:
            folder (Path): The output folder from an Experiment, from which we load
                all required elements to make a inference pipeline.
            serialiser (Optional[Serialiser]): The Serialiser used to deserialise the pipeline and input/output
                type checkers. Defaults to a JoblibSerialiser.

        Returns:
            Inferencer: The inference pipeline that can predict for new data.
        """
        folder = folder.absolute()
        with open(folder / Constants.EXTRA_FILES, "r") as f:
            extra_files = json.load(f)["paths"]

        with switch_workspace(folder, extra_files):
            if serialiser is None:
                tmp_serialiser = JoblibSerialiser()
                serialiser = tmp_serialiser.deserialise(folder / Constants.SERIALISER_FILE)
            assert isinstance(serialiser, Serialiser), "Please provide a Serialiser!"

            model = Model.load(folder / Constants.MODEL_FOLDER)
            pipeline = serialiser.deserialise(folder / Constants.PIPELINE_FILE)
            input_type_checker = serialiser.deserialise(folder / Constants.INPUT_TYPE_CHECKER_FILE)
            output_type_checker = serialiser.deserialise(folder / Constants.OUTPUT_TYPE_CHECKER_FILE)

        return cls(
            model=model,
            pipeline=pipeline,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
        )

    @classmethod
    def from_experiment(cls, experiment: Experiment) -> "Inferencer":
        """Generates an Inferencer from an experiment.

        We currently cannot check if the experiment has been trained already.

        Args:
            experiment (Experiment): The experiment to convert to a inferencing
                pipeline.

        Returns:
            Inferencer: An inferencing pipeline obtained from the given experiment.
        """
        return cls(
            model=experiment.model,
            pipeline=experiment.pipeline,
            input_type_checker=experiment.input_type_checker,
            output_type_checker=experiment.output_type_checker,
        )

    def __str__(self) -> str:
        """Create string representation of this Inferencer.

        Returns:
            str: A string representation of this Inferencer.
        """
        pipeline_part = "pipeline:\n" + self.pipeline.__str__(indents=1)
        model_part = f"model:\n\t{str(self.model)}"

        inputs = f"inputs:\n\t{str(self.input_type_checker)}"
        outputs = f"outputs:\n\t{str(self.output_type_checker)}"

        return "\n\n".join([inputs, outputs, pipeline_part, model_part])
